<?xml version="1.0" encoding="US-ASCII"?>
<?xml-stylesheet type="text/xsl" href="../xslstyle/xslstyle-docbook.xsl"?>
<xsl:stylesheet xmlns:xsl="http://www.w3.org/1999/XSL/Transform"
                xmlns:xs="http://www.CraneSoftwrights.com/ns/xslstyle"
                xmlns:xsd="http://www.w3.org/2001/XMLSchema"
                xmlns:c="http://CraneSoftwrights.com/ns/parseCSV"
                exclude-result-prefixes="xs xsd c"
                version="2.0">

<xs:doc info="$Id: Crane-ParseCSV.xsl,v 1.9 2013/01/18 22:30:24 admin Exp $"
        filename="parseCSV.xsl" vocabulary="DocBook">
  <xs:title>Read a text file into records</xs:title>
  <para>
    This stylesheet defines a named template that can be invoked to translate
    CSV content into a set of <literal>&lt;record></literal> elements.
  </para>
</xs:doc>

<xs:template>
  <para>
    Walk through the given CSV data or file, creating a record for each
    line and a field for each comma-separated field:
  </para>
<programlisting><![CDATA[
  <record>
    <field>...</field>
    <field>...</field>
    <field>...</field>
  </record>
  <record>
    <field>...</field>
    <field>...</field>
    <field>...</field>
  </record>
  <record>
    <field>...</field>
    <field>...</field>
    <field>...</field>
  </record>
]]></programlisting>
  <para>
    This assumes any field with a comma, quote or end-of-line sequence is
    quoted and an embedded quote is escaped using two quotes.
  </para>
  <para>
    Ref: <ulink url="http://en.wikipedia.org/wiki/Comma-separated_values"
    ><literal>http://en.wikipedia.org/wiki/Comma-separated_values</literal
    ></ulink>
  </para>
  <xs:param name="csv">
    <para>
      The comma-separated values content to be parsed into records.
    </para>
  </xs:param>
  <xs:param name="filename">
    <para>
      The name of a CSV file to open when CSV content is not supplied.  When
      it is a relative URI it is resolved relative to the URI of this
      stylesheet.  This argument is ignored if CSV content is explicitly
      supplied.
    </para>
  </xs:param>
  <xs:param name="encoding">
    <para>
      The encoding of the input file, defaulting to UTF-8.
    </para>
  </xs:param>
  <xs:param name="xlate-eol">
    <para>
      The indication all quoted end-of-line sequences should be translated to
      a single space.  The default is to preserve end-of-line sequences as
      a single newline character.
    </para>
  </xs:param>
</xs:template>
<xsl:template name="c:parseCSV">
  <xsl:param name="filename" as="xsd:string?"/>
  <xsl:param name="encoding" as="xsd:string" select="'UTF-8'"/>
  <xsl:param name="csv" as="xsd:string?">
    <xsl:if test="not($filename)">
      <xsl:message terminate="yes">
        <xsl:text>No CSV content or file name supplied</xsl:text>
      </xsl:message>
    </xsl:if>
    <xsl:if 
    test="not(unparsed-text-available(resolve-uri($filename,static-base-uri()),
                                      $encoding))">
      <xsl:message terminate="yes">
        <xsl:text>Problem opening or accessing the input file: </xsl:text>
        <xsl:value-of select="resolve-uri($filename,static-base-uri())"/>
      </xsl:message>
    </xsl:if>
  <xsl:value-of select="unparsed-text(resolve-uri($filename,static-base-uri()),
                                                  $encoding)"/>
  </xsl:param>
  <xsl:param name="xlate-eol" as="xsd:boolean" select="false()"/>
  <!--act on the entire file-->
  <xsl:analyze-string flags="s"
      select="concat($csv,if( ends-with($csv,'&#xa;') ) then '' else '&#xa;')
              (:guarantee last match:)"
      regex='(((("((([^"]*)("")?)*)")|[^",\r\n]*)(,?))*)(\r?\n|\r)'>
    <xsl:matching-substring>
      <!--parse a record that terminates with a newline-->
      <xsl:call-template name="c:parseRecord">
        <xsl:with-param name="record" select="regex-group(1)"/>
        <xsl:with-param name="xlate-eol" select="$xlate-eol"/>
      </xsl:call-template>
    </xsl:matching-substring>
  </xsl:analyze-string>
</xsl:template>

<xs:template>
  <para>Parse a record full of fields.</para>
  <xs:param name="record">
    <para>
      The text of a record of fields.
    </para>
  </xs:param>
  <xs:param name="xlate-eol">
    <para>
      An indication if any end-of-line sequences should be translated to
      a single space.
    </para>
  </xs:param>
  <xs:param name="debug-field-number">
    <para>
      When non-zero, this will expose the running record number and the given
      field number for each record.
    </para>
  </xs:param>
</xs:template>
<xsl:template name="c:parseRecord">
  <xsl:param name="record" as="xsd:string?"/>
  <xsl:param name="xlate-eol" as="xsd:boolean"/>
  <xsl:param name="debug-field-number" as="xsd:integer" select="0"
             tunnel="yes"/>
  <xsl:variable name="recordNumber" select="position()"/>
  <record>
    <xsl:if test="$record">
      <!--loop across the fields of a record-->
      <xsl:analyze-string flags="s"
                          select="concat($record,',')(:guarantee last match:)"
                          regex='(("((([^"]*)("")?)*)")|([^",]*)),'>
        <!--found a field-->
        <xsl:matching-substring>
          <!--parse a record that terminates with a newline-->
          <xsl:if test="position()=$debug-field-number">
            <xsl:message select="'Record:',$recordNumber,."/>
          </xsl:if>
          <xsl:call-template name="c:parseField">
            <xsl:with-param name="field" select="regex-group(1)"/>
            <xsl:with-param name="xlate-eol" select="$xlate-eol"/>
          </xsl:call-template>
        </xsl:matching-substring>
      </xsl:analyze-string>
    </xsl:if>
  </record>
</xsl:template>

<xs:template>
  <para>
    Parse a field value.  If the first or last character of the field is a
    quote, then it must have been a quoted field.
  </para>
  <xs:param name="field">
    <para>
      The text of a record of fields.
    </para>
  </xs:param>
  <xs:param name="xlate-eol">
    <para>
      An indication if any end-of-line sequences should be translated to
      a single space.
    </para>
  </xs:param>
</xs:template>
<xsl:template name="c:parseField">
  <xsl:param name="field" as="xsd:string?"/>
  <xsl:param name="xlate-eol" as="xsd:boolean"/>
  <field>
    <xsl:value-of select='replace(replace(replace(replace($field,"^""",""),
                                                  """$",""),
                                          """""",""""),
                            "(\n)",if( $xlate-eol ) then " " else "$1","s")'/>
  </field>
</xsl:template>

</xsl:stylesheet>
